@@ -1235,7 +1235,7 @@ double edge_distance_to_point(const GEOGRAPHIC_EDGE *e, const GEOGRAPHIC_POINT *
 	vector_difference(&p, &n, &k);
 	normalize(&k);
 	cart2geog(&k, &gk);
-	if ( edge_point_in_cone(e, &gk) )
+	if ( edge_contains_point(e, &gk) )
 	{
 		d1 = sphere_distance(gp, &gk);
 	}
@@ -2156,7 +2156,7 @@ LWPOINT* lwgeom_project_spheroid(const LWPOINT *r, const SPHEROID *spheroid, dou
 double lwgeom_azumith_spheroid(const LWPOINT *r, const LWPOINT *s, const SPHEROID *spheroid)
 {
 	GEOGRAPHIC_POINT g1, g2;
-	double x1, y1, x2, y2;
+	double x1, y1, x2, y2, az;
 
 	/* Convert r to a geodetic point */
 	x1 = lwpoint_get_x(r);
@@ -2175,7 +2175,10 @@ double lwgeom_azumith_spheroid(const LWPOINT *r, const LWPOINT *s, const SPHEROI
 	}
 
 	/* Do the direction calculation */
-	return spheroid_direction(&g1, &g2, spheroid);
+	az = spheroid_direction(&g1, &g2, spheroid);
+	/* Ensure result is positive */
+	return az < -0 ? 2*M_PI + az : az;
+	// return az;
 }
 
 /**
@@ -2992,7 +2995,7 @@ static int lwtriangle_calculate_gbox_geodetic(const LWTRIANGLE *triangle, GBOX *
 
 static int lwcollection_calculate_gbox_geodetic(const LWCOLLECTION *coll, GBOX *gbox)
 {
-	GBOX subbox;
+	GBOX subbox = {0};
 	uint32_t i;
 	int result = LW_FAILURE;
 	int first = LW_TRUE;
